	list n=0	; Remove page breaks in .lst file

; *******************************************************************************
; *										*
; *	Filename: VK5TM_16F628_LCD_I2C_Backpack.asm				*
; *	Author: Terry Mowles VK5TM (C)2023					*
; *	Date: Dec 2023								*
; *										*
; *	16F628A bit-banged I2C demo for PCF8574 I2C LCD Backpack and		*
; *	HD44780 type LCD display.						*
; *										*
; *	This code is for single i2c device connection only			*
; *										*
; *	Parts of the code by Mike McLaren, K8LH					*
; *										*
;********************************************************************************
        list P=PIC16F628A 

       #include <P16F628A.INC>

        errorlevel -302
        radix dec

  	__config _CP_OFF&_LVP_OFF&_BODEN_OFF&_MCLRE_OFF&_PWRTE_ON&_WDT_OFF&_INTOSC_OSC_NOCLKOUT
;
; *******************************************************************************
; *	Defines									*
; *******************************************************************************
;
SDA	equ	0x07		; I2C Data
SCL	equ	0x06		; I2C Clock
;
;  LCD DDRAM address constants
;
line1   equ     127+0           ; LCD "line 1" command
line2   equ     127+64          ; LCD "line 2" command
;
;********************************************************************************
; *	Variables								*
; *******************************************************************************
        cblock  0x20
	delay1
	delay2
	delay3
	bytcnt
	rsflag
	workvar
	w_data
        endc
 ;
; *******************************************************************************
; *	Start									*
; *******************************************************************************
	org	0x000

	clrf	STATUS          ; force bank 0 and IRP = 0
	goto	setup           ;

	org	0x004
	goto	setup           ;
;
; *******************************************************************************
; *	main setup								*
; *******************************************************************************
;
setup
	banksel TRISA 
 	movlw	0xFF
	movwf	TRISA		; all inputs
	bcf	TRISA,SDA	; Make SDA & SCL outputs
	bcf	TRISA,SCL
	clrf	TRISB		; PortB all outputs
	banksel	0
	clrf	PORTA		;
	clrf	PORTB		;

	bsf	PORTA,SDA	; i2c idle state = SDA & SCL high
	bsf	PORTA,SCL

	movlw	7
	movwf	CMCON
;
;  LCD "initialize by instruction" procedure for 4-bit interface K8LH
;
	movlw	30		; delay 30-msecs after power up
	call	w_mS
	movlw	0x30		; function set: 8-bit
	call	lcdNyb
	movlw	5		; required 5-msec delay
	call	w_mS
	movlw	0x30		; function set: 8-bit
	call	lcdNyb
	movlw	1		; 
	call	w_mS
	movlw	0x30		; function set: 8-bit
	call	lcdNyb
	movlw	1		; 
	call	w_mS
;
;  now we're in 4-bit mode and can handle 8-bit transactions by
;  sending both hi & lo nibbles using lcdCmd & lcdDat.
;
	movlw	0x28		; 4-bit, 2-lines, 5x7 font
	call	lcdCmd
	movlw	1		; 
	call	w_mS
	movlw	0x08		; display, cursor, blink all off
	call	lcdCmd
	movlw	1		; 
	call	w_mS
	movlw	0x01		; clear display
	call	lcdCmd
	movlw	2		; required delay
	call	w_mS
 	movlw	0x06		; cursor inc, shift off
	call	lcdCmd
	movlw	1		; 
	call	w_mS
	movlw	0x0C		; display on, leave cursor off
	call	lcdCmd

;******************************************************************
;  main loop                                                      *
;******************************************************************

loop
        movlw	line1+3		; LCD line 1, column 3
	call	lcdCmd
        movlw	'H'
	call	lcdDat
        movlw	'e'
	call	lcdDat
        movlw	'l'
	call	lcdDat
        movlw	'l'
	call	lcdDat
        movlw	'o'
	call	lcdDat
        movlw	' '
	call	lcdDat
        movlw	'W'
	call	lcdDat
        movlw	'o'
	call	lcdDat
        movlw	'r'
	call	lcdDat
        movlw	'l'
	call	lcdDat
        movlw	'd'
	call	lcdDat
        movlw	'!'
	call	lcdDat

        goto    $

;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;  low-level LCD drivers for PCF8574 based I2C LCD backpack K8LH  ~
;~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

lcdCmd                          ; entry point for "cmd" data
        clrf    rsflag          ; wreg = data, RS = 0 (cmd)       |B0
	btfss	STATUS,Z        ; skip 
lcdDat                          ; entry point for "dat" data
        bsf     rsflag,0        ; wreg = data, RS = 1 (data)      |B0
        call    lcdNyb          ; send hi nybble                  |B0
        swapf   workvar,W       ; send lo nybble                  |B0
;
;  ~~ 'lcdNyb' sequence ~~~~~~~~~      ~~< PCF8574 format >~~~
;  () I2C 'start' + I2C 'address'      P0(b0) -> LCD 'RS'
;  () write nibble, E = 1              P1(b1) -> LCD 'RW' (0)
;  () write nibble, E = 0              P2(b2) -> LCD 'E'
;  () I2C 'stop'                       P3(b3) -> LCD Backlight
;                                      P4(b4) -> LCD 'D4'
;   ~617-usecs (isochronous) for       P5(b5) -> LCD 'D5'
;   each lcdCmd and lcdDat call.       P6(b6) -> LCD 'D6'
;                                      P7(b7) -> LCD 'D7'
lcdNyb
	movwf	w_data		; save temporarily                |B0
 ;       movlw   0x4E            ; PCF8574 I2C address + 0 (wr) moved to i2c_start|B0 
	call	i2c_start	; I2C 'start' + 'address' + 'rw'  |B0
	movf	w_data,W	;                                 |B0
	andlw	0xF0		; use left nybble (b7..b4) bits   |B0
	btfsc	rsflag,0	; RS = 0? yes, skip, else         |B0
	iorlw	b'00000001'	; RS(b0) = 1                      |B0
	iorlw	b'00001100'	; Backlight(b3) = 1, E(b2) = 1    |B0
	call	i2c_write	; data + bl, en, rw, rs bits      |B0
	xorlw	b'00000100'	; clr E bit (b2)                  |B0
	call	i2c_write	; data + bl, en, rw, rs bits      |B0
	call	i2c_stop	; I2C 'stop'                      |B0

	return
;
; *******************************************************************************
; *				I2C ROUTINES					*
; *******************************************************************************
; Start condition; SDA goes from hi to low with SCL high
; Normal state is SDA high, SCL high
i2c_start
	bsf	PORTA,SDA	; Set SDA High	ensure idle state
	nop
	bsf	PORTA,SCL	; Set SCL hi
	goto	$+1
	goto	$+1		;\
	goto	$+1		; To set timing 4MHz clock
	goto	$+1		;/
	goto	$+1
	bcf	PORTA,SDA	; SDA low
	nop
	bcf	PORTA,SCL	; SCL low

	movlw	0x4E		; Send i2c address
;
; *******************************************************************************	
; *	Output a byte in workvar and then get an ACK				*
; *******************************************************************************
;
i2c_write
	movwf	workvar

	clrf	bytcnt	
	bsf	bytcnt,3	; byte counter = 8 (preserve w)

	bcf	STATUS,C
send_data
	rlf	workvar,f
	btfss	STATUS,C
	goto	bit_0			
	bsf	PORTA,SDA	; SDA hi
	nop
	goto	bit_1
bit_0
	bcf	PORTA,SDA	; SDA low
	nop
bit_1
	bsf	PORTA,SCL	; SCL hi
	goto	$+1
	goto	$+1
	goto	$+1
	goto	$+1
	goto	$+1
	bcf	PORTA,SCL	; SCL low
	goto	$+1
	nop
	decfsz	bytcnt,f			
	goto	send_data
;
; *******************************************************************************
; *	Read acknowledge signal.						*
; *******************************************************************************
;
getack
	banksel	TRISA	
	bsf	TRISA,SDA	; SDA is Input
	banksel 0	
	bsf	PORTA,SCL	; SCL hi  
	goto	$+1
	goto	$+1
	goto	$+1
	goto	$+1
	goto	$+1
	bcf	PORTA,SCL	; SCL low 
	nop
	banksel	TRISA	
	bcf	TRISA,SDA	; SDA is output
	banksel	0	

	return
;
; *******************************************************************************
; *	Stop condition; SDA goes from low to hi with SCL high			*
; *	Normal state is SDA high SCL high					*
; *******************************************************************************
;
i2c_stop

	bcf	PORTA,SDA	; SDA low 
	nop
	bsf	PORTA,SCL	; SCL hi  
	goto	$+1
	goto	$+1
	goto	$+1
	goto	$+1
	goto	$+1
	bsf	PORTA,SDA	; SDA hi 

	return
;
; *******************************************************************************
; *	DELAY LOOP w x milliseconds - not optimised				*
; *******************************************************************************
;
w_mS
	movwf	delay3		; Enter with 'w' holding required mS
w_mS_loop
	call	one_mS
	decfsz	delay3,f
	goto	w_mS_loop
	return
;
; DELAY LOOP 1 millisecond @ 4MHz
;
one_mS
	movlw	d'4'
	movwf	delay1
loop1
	movlw	d'81'
	movwf	delay2
loop2
	decfsz	delay2,f
	goto	loop2
	decfsz	delay1,f
	goto	loop1
	goto	$+1
	goto	$+1
	goto	$+1
	return

        end